from dagster import check
from dagster.core.definitions.run_request import InstigatorType
from dagster.core.host_representation import PipelineSelector, RepositorySelector, ScheduleSelector
from dagster.core.scheduler.instigation import InstigatorStatus
from dagster.seven import get_current_datetime_in_utc, get_timestamp_from_utc_datetime
from graphql.execution.base import ResolveInfo

from .utils import UserFacingGraphQLError, capture_error


@capture_error
def start_schedule(graphene_info, schedule_selector):
    from ..schema.instigation import GrapheneInstigationState
    from ..schema.schedules import GrapheneScheduleStateResult

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(schedule_selector, "schedule_selector", ScheduleSelector)
    location = graphene_info.context.get_repository_location(schedule_selector.location_name)
    repository = location.get_repository(schedule_selector.repository_name)
    instance = graphene_info.context.instance
    schedule_state = instance.start_schedule_and_update_storage_state(
        repository.get_external_schedule(schedule_selector.schedule_name)
    )
    return GrapheneScheduleStateResult(GrapheneInstigationState(schedule_state))


@capture_error
def stop_schedule(graphene_info, schedule_origin_id):
    from ..schema.instigation import GrapheneInstigationState
    from ..schema.schedules import GrapheneScheduleStateResult

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    instance = graphene_info.context.instance
    schedule_state = instance.stop_schedule_and_update_storage_state(schedule_origin_id)
    return GrapheneScheduleStateResult(GrapheneInstigationState(schedule_state))


@capture_error
def get_scheduler_or_error(graphene_info):
    from ..schema.errors import GrapheneSchedulerNotDefinedError
    from ..schema.schedules import GrapheneScheduler

    instance = graphene_info.context.instance

    if not instance.scheduler:
        raise UserFacingGraphQLError(GrapheneSchedulerNotDefinedError())

    return GrapheneScheduler(scheduler_class=instance.scheduler.__class__.__name__)


@capture_error
def get_schedules_or_error(graphene_info, repository_selector):
    from ..schema.schedules import GrapheneSchedule, GrapheneSchedules

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(repository_selector, "repository_selector", RepositorySelector)

    location = graphene_info.context.get_repository_location(repository_selector.location_name)
    repository = location.get_repository(repository_selector.repository_name)
    external_schedules = repository.get_external_schedules()
    schedule_states_by_name = {
        state.name: state
        for state in graphene_info.context.instance.all_stored_job_state(
            repository_origin_id=repository.get_external_origin_id(),
            job_type=InstigatorType.SCHEDULE,
        )
    }

    results = [
        GrapheneSchedule(
            external_schedule,
            schedule_states_by_name.get(external_schedule.name),
        )
        for external_schedule in external_schedules
    ]

    return GrapheneSchedules(results=results)


def get_schedules_for_pipeline(graphene_info, pipeline_selector):
    from ..schema.schedules import GrapheneSchedule

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(pipeline_selector, "pipeline_selector", PipelineSelector)

    location = graphene_info.context.get_repository_location(pipeline_selector.location_name)
    repository = location.get_repository(pipeline_selector.repository_name)
    external_schedules = repository.get_external_schedules()

    results = []
    for external_schedule in external_schedules:
        if external_schedule.pipeline_name != pipeline_selector.pipeline_name:
            continue

        schedule_state = graphene_info.context.instance.get_job_state(
            external_schedule.get_external_origin_id()
        )
        results.append(GrapheneSchedule(external_schedule, schedule_state))

    return results


@capture_error
def get_schedule_or_error(graphene_info, schedule_selector):
    from ..schema.errors import GrapheneScheduleNotFoundError
    from ..schema.schedules import GrapheneSchedule

    check.inst_param(graphene_info, "graphene_info", ResolveInfo)
    check.inst_param(schedule_selector, "schedule_selector", ScheduleSelector)
    location = graphene_info.context.get_repository_location(schedule_selector.location_name)
    repository = location.get_repository(schedule_selector.repository_name)

    external_schedule = repository.get_external_schedule(schedule_selector.schedule_name)
    if not external_schedule:
        raise UserFacingGraphQLError(
            GrapheneScheduleNotFoundError(schedule_name=schedule_selector.schedule_name)
        )

    schedule_state = graphene_info.context.instance.get_job_state(
        external_schedule.get_external_origin_id()
    )
    return GrapheneSchedule(external_schedule, schedule_state)


def get_schedule_next_tick(graphene_info, schedule_state):
    from ..schema.instigation import GrapheneFutureInstigationTick

    if schedule_state.status != InstigatorStatus.RUNNING:
        return None

    repository_origin = schedule_state.origin.external_repository_origin
    if not graphene_info.context.has_repository_location(
        repository_origin.repository_location_origin.location_name
    ):
        return None
    repository_location = graphene_info.context.get_repository_location(
        repository_origin.repository_location_origin.location_name
    )
    if not repository_location.has_repository(repository_origin.repository_name):
        return None

    repository = repository_location.get_repository(repository_origin.repository_name)
    external_schedule = repository.get_external_schedule(schedule_state.name)
    time_iter = external_schedule.execution_time_iterator(
        get_timestamp_from_utc_datetime(get_current_datetime_in_utc())
    )

    next_timestamp = next(time_iter).timestamp()
    return GrapheneFutureInstigationTick(schedule_state, next_timestamp)



import sys
import tempfile
from abc import ABC, abstractmethod
from contextlib import contextmanager
from unittest.mock import patch

import pytest
from dagster import check, file_relative_path
from dagster.core.instance import DagsterInstance, InstanceType
from dagster.core.launcher.sync_in_memory_run_launcher import SyncInMemoryRunLauncher
from dagster.core.run_coordinator import DefaultRunCoordinator
from dagster.core.storage.event_log import InMemoryEventLogStorage
from dagster.core.storage.event_log.sqlite import ConsolidatedSqliteEventLogStorage
from dagster.core.storage.local_compute_log_manager import LocalComputeLogManager
from dagster.core.storage.root import LocalArtifactStorage
from dagster.core.storage.runs import InMemoryRunStorage
from dagster.core.storage.schedules.sqlite.sqlite_schedule_storage import SqliteScheduleStorage
from dagster.core.test_utils import ExplodingRunLauncher, instance_for_test
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.core.workspace import WorkspaceProcessContext
from dagster.core.workspace.load_target import (
    GrpcServerTarget,
    ModuleTarget,
    PythonFileTarget,
    WorkspaceFileTarget,
)
from dagster.grpc.server import GrpcServerProcess
from dagster.utils import merge_dicts
from dagster.utils.test import FilesystemTestScheduler
from dagster.utils.test.postgres_instance import TestPostgresInstance
from dagster_graphql import DagsterGraphQLClient
from dagster_graphql.test.utils import execute_dagster_graphql


def get_main_loadable_target_origin():
    return LoadableTargetOrigin(
        executable_path=sys.executable,
        python_file=file_relative_path(__file__, "setup.py"),
        attribute="test_repo",
    )


@contextmanager
def graphql_postgres_instance(overrides):
    with tempfile.TemporaryDirectory() as temp_dir:
        with TestPostgresInstance.docker_service_up_or_skip(
            file_relative_path(__file__, "docker-compose.yml"),
            "test-postgres-db-graphql",
        ) as pg_conn_string:
            TestPostgresInstance.clean_run_storage(pg_conn_string)
            TestPostgresInstance.clean_event_log_storage(pg_conn_string)
            TestPostgresInstance.clean_schedule_storage(pg_conn_string)
            with instance_for_test(
                temp_dir=temp_dir,
                overrides=merge_dicts(
                    {
                        "run_storage": {
                            "module": "dagster_postgres.run_storage.run_storage",
                            "class": "PostgresRunStorage",
                            "config": {"postgres_url": pg_conn_string},
                        },
                        "event_log_storage": {
                            "module": "dagster_postgres.event_log.event_log",
                            "class": "PostgresEventLogStorage",
                            "config": {"postgres_url": pg_conn_string},
                        },
                        "schedule_storage": {
                            "module": "dagster_postgres.schedule_storage.schedule_storage",
                            "class": "PostgresScheduleStorage",
                            "config": {"postgres_url": pg_conn_string},
                        },
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                    },
                    overrides if overrides else {},
                ),
            ) as instance:
                yield instance


class MarkedManager:
    """
    MarkedManagers are passed to GraphQLContextVariants. They contain
    a contextmanager function "manager_fn" that yield the relevant
    instace, and it includes marks that will be applied to any
    context-variant-driven test case that includes this MarkedManager.

    See InstanceManagers for an example construction.

    See GraphQLContextVariant for further information
    """

    def __init__(self, manager_fn, marks):
        self.manager_fn = check.callable_param(manager_fn, "manager_fn")
        self.marks = check.list_param(marks, "marks")


class InstanceManagers:
    @staticmethod
    def in_memory_instance():
        @contextmanager
        def _in_memory_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                yield DagsterInstance(
                    instance_type=InstanceType.EPHEMERAL,
                    local_artifact_storage=LocalArtifactStorage(temp_dir),
                    run_storage=InMemoryRunStorage(),
                    event_storage=InMemoryEventLogStorage(),
                    compute_log_manager=LocalComputeLogManager(temp_dir),
                    run_launcher=SyncInMemoryRunLauncher(),
                    run_coordinator=DefaultRunCoordinator(),
                    schedule_storage=SqliteScheduleStorage.from_local(temp_dir),
                    scheduler=FilesystemTestScheduler(temp_dir),
                )

        return MarkedManager(_in_memory_instance, [Marks.in_memory_instance])

    @staticmethod
    def non_launchable_in_memory_instance():
        @contextmanager
        def _non_launchable_in_memory_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                yield DagsterInstance(
                    instance_type=InstanceType.EPHEMERAL,
                    local_artifact_storage=LocalArtifactStorage(temp_dir),
                    run_storage=InMemoryRunStorage(),
                    event_storage=InMemoryEventLogStorage(),
                    compute_log_manager=LocalComputeLogManager(temp_dir),
                    run_launcher=ExplodingRunLauncher(),
                    run_coordinator=DefaultRunCoordinator(),
                    schedule_storage=SqliteScheduleStorage.from_local(temp_dir),
                    scheduler=FilesystemTestScheduler(temp_dir),
                )

        return MarkedManager(
            _non_launchable_in_memory_instance,
            [Marks.in_memory_instance, Marks.non_launchable],
        )

    @staticmethod
    def non_launchable_sqlite_instance():
        @contextmanager
        def _non_launchable_sqlite_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                        "run_launcher": {
                            "module": "dagster.core.test_utils",
                            "class": "ExplodingRunLauncher",
                        },
                    },
                ) as instance:
                    yield instance

        return MarkedManager(
            _non_launchable_sqlite_instance, [Marks.sqlite_instance, Marks.non_launchable]
        )

    @staticmethod
    def non_launchable_postgres_instance():
        @contextmanager
        def _non_launchable_postgres_instance():
            with graphql_postgres_instance(
                overrides={
                    "run_launcher": {
                        "module": "dagster.core.test_utils",
                        "class": "ExplodingRunLauncher",
                    }
                }
            ) as instance:
                yield instance

        return MarkedManager(
            _non_launchable_postgres_instance,
            [Marks.postgres_instance, Marks.non_launchable],
        )

    @staticmethod
    def sqlite_instance_with_sync_run_launcher():
        @contextmanager
        def _sqlite_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                        "run_launcher": {
                            "module": "dagster.core.launcher.sync_in_memory_run_launcher",
                            "class": "SyncInMemoryRunLauncher",
                        },
                    },
                ) as instance:
                    yield instance

        return MarkedManager(_sqlite_instance, [Marks.sqlite_instance, Marks.sync_run_launcher])

    # Runs launched with this instance won't actually execute since the graphql test suite
    # doesn't run the daemon process that launches queued runs
    @staticmethod
    def sqlite_instance_with_queued_run_coordinator():
        @contextmanager
        def _sqlite_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                        "run_coordinator": {
                            "module": "dagster.core.run_coordinator.queued_run_coordinator",
                            "class": "QueuedRunCoordinator",
                        },
                    },
                ) as instance:
                    yield instance

        return MarkedManager(
            _sqlite_instance, [Marks.sqlite_instance, Marks.queued_run_coordinator]
        )

    @staticmethod
    def sqlite_instance_with_default_run_launcher():
        @contextmanager
        def _sqlite_instance_with_default_hijack():
            with tempfile.TemporaryDirectory() as temp_dir:
                with instance_for_test(
                    temp_dir=temp_dir,
                    overrides={
                        "scheduler": {
                            "module": "dagster.utils.test",
                            "class": "FilesystemTestScheduler",
                            "config": {"base_dir": temp_dir},
                        },
                        "run_launcher": {
                            "module": "dagster",
                            "class": "DefaultRunLauncher",
                            "config": {
                                "wait_for_processes": True,
                            },
                        },
                    },
                ) as instance:
                    yield instance

        return MarkedManager(
            _sqlite_instance_with_default_hijack,
            [Marks.sqlite_instance, Marks.default_run_launcher],
        )

    @staticmethod
    def postgres_instance_with_sync_run_launcher():
        @contextmanager
        def _postgres_instance():
            with graphql_postgres_instance(
                overrides={
                    "run_launcher": {
                        "module": "dagster.core.launcher.sync_in_memory_run_launcher",
                        "class": "SyncInMemoryRunLauncher",
                    }
                }
            ) as instance:
                yield instance

        return MarkedManager(
            _postgres_instance,
            [Marks.postgres_instance, Marks.sync_run_launcher],
        )

    @staticmethod
    def postgres_instance_with_default_run_launcher():
        @contextmanager
        def _postgres_instance_with_default_hijack():
            with graphql_postgres_instance(
                overrides={
                    "run_launcher": {
                        "module": "dagster",
                        "class": "DefaultRunLauncher",
                        "config": {
                            "wait_for_processes": True,
                        },
                    },
                }
            ) as instance:
                yield instance

        return MarkedManager(
            _postgres_instance_with_default_hijack,
            [Marks.postgres_instance, Marks.default_run_launcher],
        )

    @staticmethod
    def consolidated_sqlite_instance():
        @contextmanager
        def _sqlite_asset_instance():
            with tempfile.TemporaryDirectory() as temp_dir:
                instance = DagsterInstance(
                    instance_type=InstanceType.EPHEMERAL,
                    local_artifact_storage=LocalArtifactStorage(temp_dir),
                    run_storage=InMemoryRunStorage(),
                    event_storage=ConsolidatedSqliteEventLogStorage(temp_dir),
                    compute_log_manager=LocalComputeLogManager(temp_dir),
                    run_coordinator=DefaultRunCoordinator(),
                    run_launcher=SyncInMemoryRunLauncher(),
                    scheduler=FilesystemTestScheduler(temp_dir),
                )
                yield instance

        return MarkedManager(_sqlite_asset_instance, [Marks.asset_aware_instance])


class EnvironmentManagers:
    @staticmethod
    def managed_grpc():
        @contextmanager
        def _mgr_fn(instance, read_only):
            """Goes out of process via grpc"""
            loadable_target_origin = get_main_loadable_target_origin()
            with WorkspaceProcessContext(
                instance,
                (
                    PythonFileTarget(
                        python_file=loadable_target_origin.python_file,
                        attribute=loadable_target_origin.attribute,
                        working_directory=loadable_target_origin.working_directory,
                        location_name="test",
                    )
                    if loadable_target_origin.python_file
                    else ModuleTarget(
                        module_name=loadable_target_origin.module_name,
                        attribute=loadable_target_origin.attribute,
                        working_directory=loadable_target_origin.working_directory,
                        location_name="test",
                    )
                ),
                version="",
                read_only=read_only,
            ) as workspace_process_context:
                yield workspace_process_context

        return MarkedManager(_mgr_fn, [Marks.managed_grpc_env])

    @staticmethod
    def deployed_grpc():
        @contextmanager
        def _mgr_fn(instance, read_only):
            server_process = GrpcServerProcess(
                loadable_target_origin=get_main_loadable_target_origin()
            )
            try:
                with server_process.create_ephemeral_client() as api_client:
                    with WorkspaceProcessContext(
                        instance,
                        GrpcServerTarget(
                            port=api_client.port,
                            socket=api_client.socket,
                            host=api_client.host,
                            location_name="test",
                        ),
                        version="",
                        read_only=read_only,
                    ) as workspace:
                        yield workspace
            finally:
                server_process.wait()

        return MarkedManager(_mgr_fn, [Marks.deployed_grpc_env])

    @staticmethod
    def multi_location():
        @contextmanager
        def _mgr_fn(instance, read_only):
            """Goes out of process but same process as host process"""
            with WorkspaceProcessContext(
                instance,
                WorkspaceFileTarget(paths=[file_relative_path(__file__, "multi_location.yaml")]),
                version="",
                read_only=read_only,
            ) as workspace:
                yield workspace

        return MarkedManager(_mgr_fn, [Marks.multi_location])

    @staticmethod
    def lazy_repository():
        @contextmanager
        def _mgr_fn(instance, read_only):
            """Goes out of process but same process as host process"""
            with WorkspaceProcessContext(
                instance,
                PythonFileTarget(
                    python_file=file_relative_path(__file__, "setup.py"),
                    attribute="test_dict_repo",
                    working_directory=None,
                    location_name="test",
                ),
                version="",
                read_only=read_only,
            ) as workspace:
                yield workspace

        return MarkedManager(_mgr_fn, [Marks.lazy_repository])


class Marks:
    # Instance type makes
    in_memory_instance = pytest.mark.in_memory_instance
    sqlite_instance = pytest.mark.sqlite_instance
    postgres_instance = pytest.mark.postgres_instance

    # Run launcher variants
    sync_run_launcher = pytest.mark.sync_run_launcher
    default_run_launcher = pytest.mark.default_run_launcher
    queued_run_coordinator = pytest.mark.queued_run_coordinator
    non_launchable = pytest.mark.non_launchable

    # Repository Location marks
    multi_location = pytest.mark.multi_location
    managed_grpc_env = pytest.mark.managed_grpc_env
    deployed_grpc_env = pytest.mark.deployed_grpc_env
    lazy_repository = pytest.mark.lazy_repository

    # Asset-aware sqlite variants
    asset_aware_instance = pytest.mark.asset_aware_instance

    # Backfill daemon variants
    backfill_daemon = pytest.mark.backfill_daemon

    # Readonly context variant
    read_only = pytest.mark.read_only

    # Common mark to all test suite tests
    graphql_context_test_suite = pytest.mark.graphql_context_test_suite


def none_manager():
    @contextmanager
    def _yield_none(*_args, **_kwargs):
        yield None

    return MarkedManager(_yield_none, [])


class GraphQLContextVariant:
    """
    An instance of this class represents a context variant that will be run
    against *every* method in the test class, defined as a class
    created by inheriting from make_graphql_context_test_suite.

    It comes with a number of static methods with prebuilt context variants.
    e.g. in_memory_in_process_start

    One can also make bespoke context variants, provided you configure it properly
    with MarkedMembers that produce its members.

    Args:

    marked_instance_mgr (MarkedManager): The manager_fn
    within it must be a contextmanager that takes zero arguments and yields
    a DagsterInstance

    See InstanceManagers for examples

    marked_environment_mgr (MarkedManager): The manager_fn with in
    must be a contextmanager takes a default ReconstructableRepo and
    yields a list of RepositoryLocation.

    See EnvironmentManagers for examples

    test_id [Optional] (str): This assigns a test_id to test parameterized with this
    variant. This is highly convenient for running a particular variant across
    the entire test suite, without running all the other variants.

    e.g.
    pytest python_modules/dagster-graphql/dagster_graphql_tests/ -s -k in_memory_in_process_start

    Will run all tests that use the in_memory_in_process_start, which will get a lot
    of code coverage while being very fast to run.

    All tests managed by this system are marked with "graphql_context_test_suite".
    """

    def __init__(self, marked_instance_mgr, marked_environment_mgr, read_only=False, test_id=None):
        self.marked_instance_mgr = check.inst_param(
            marked_instance_mgr, "marked_instance_mgr", MarkedManager
        )
        self.marked_environment_mgr = check.inst_param(
            marked_environment_mgr, "marked_environment_mgr", MarkedManager
        )
        self.read_only = check.bool_param(read_only, "read_only")
        self.test_id = check.opt_str_param(test_id, "test_id")
        self.marks = (
            marked_instance_mgr.marks
            + marked_environment_mgr.marks
            + ([Marks.read_only] if read_only else [])
        )

    @property
    def instance_mgr(self):
        return self.marked_instance_mgr.manager_fn

    @property
    def environment_mgr(self):
        return self.marked_environment_mgr.manager_fn

    @staticmethod
    def in_memory_instance_managed_grpc_env():
        """
        Good for tests with read-only metadata queries. Does not work
        if you have to go through the run launcher.
        """
        return GraphQLContextVariant(
            InstanceManagers.in_memory_instance(),
            EnvironmentManagers.managed_grpc(),
            test_id="in_memory_instance_managed_grpc_env",
        )

    @staticmethod
    def sqlite_with_queued_run_coordinator_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_queued_run_coordinator(),
            EnvironmentManagers.managed_grpc(),
            test_id="sqlite_with_queued_run_coordinator_managed_grpc_env",
        )

    @staticmethod
    def sqlite_with_default_run_launcher_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_default_run_launcher(),
            EnvironmentManagers.managed_grpc(),
            test_id="sqlite_with_default_run_launcher_managed_grpc_env",
        )

    @staticmethod
    def sqlite_read_only_with_default_run_launcher_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_default_run_launcher(),
            EnvironmentManagers.managed_grpc(),
            read_only=True,
            test_id="sqlite_read_only_with_default_run_launcher_managed_grpc_env",
        )

    @staticmethod
    def sqlite_with_default_run_launcher_deployed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.sqlite_instance_with_default_run_launcher(),
            EnvironmentManagers.deployed_grpc(),
            test_id="sqlite_with_default_run_launcher_deployed_grpc_env",
        )

    @staticmethod
    def postgres_with_default_run_launcher_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.postgres_instance_with_default_run_launcher(),
            EnvironmentManagers.managed_grpc(),
            test_id="postgres_with_default_run_launcher_managed_grpc_env",
        )

    @staticmethod
    def postgres_with_default_run_launcher_deployed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.postgres_instance_with_default_run_launcher(),
            EnvironmentManagers.deployed_grpc(),
            test_id="postgres_with_default_run_launcher_deployed_grpc_env",
        )

    @staticmethod
    def non_launchable_sqlite_instance_multi_location():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_sqlite_instance(),
            EnvironmentManagers.multi_location(),
            test_id="non_launchable_sqlite_instance_multi_location",
        )

    @staticmethod
    def non_launchable_sqlite_instance_lazy_repository():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_sqlite_instance(),
            EnvironmentManagers.lazy_repository(),
            test_id="non_launchable_sqlite_instance_lazy_repository",
        )

    @staticmethod
    def non_launchable_sqlite_instance_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_sqlite_instance(),
            EnvironmentManagers.managed_grpc(),
            test_id="non_launchable_sqlite_instance_managed_grpc_env",
        )

    @staticmethod
    def non_launchable_sqlite_instance_deployed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_sqlite_instance(),
            EnvironmentManagers.deployed_grpc(),
            test_id="non_launchable_sqlite_instance_deployed_grpc_env",
        )

    @staticmethod
    def non_launchable_postgres_instance_multi_location():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_postgres_instance(),
            EnvironmentManagers.multi_location(),
            test_id="non_launchable_postgres_instance_multi_location",
        )

    @staticmethod
    def non_launchable_postgres_instance_lazy_repository():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_postgres_instance(),
            EnvironmentManagers.lazy_repository(),
            test_id="non_launchable_postgres_instance_lazy_repository",
        )

    @staticmethod
    def non_launchable_postgres_instance_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_postgres_instance(),
            EnvironmentManagers.managed_grpc(),
            test_id="non_launchable_postgres_instance_managed_grpc_env",
        )

    @staticmethod
    def non_launchable_in_memory_instance_multi_location():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_in_memory_instance(),
            EnvironmentManagers.multi_location(),
            test_id="non_launchable_in_memory_instance_multi_location",
        )

    @staticmethod
    def non_launchable_in_memory_instance_lazy_repository():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_in_memory_instance(),
            EnvironmentManagers.lazy_repository(),
            test_id="non_launchable_in_memory_instance_lazy_repository",
        )

    @staticmethod
    def non_launchable_in_memory_instance_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.non_launchable_in_memory_instance(),
            EnvironmentManagers.managed_grpc(),
            test_id="non_launchable_in_memory_instance_managed_grpc_env",
        )

    @staticmethod
    def consolidated_sqlite_instance_managed_grpc_env():
        return GraphQLContextVariant(
            InstanceManagers.consolidated_sqlite_instance(),
            EnvironmentManagers.managed_grpc(),
            test_id="asset_aware_instance_in_process_env",
        )

    @staticmethod
    def all_variants():
        """
        There is a test case that keeps this up-to-date. If you add a static
        method that returns a GraphQLContextVariant you have to add it to this
        list in order for tests to pass.
        """
        return [
            GraphQLContextVariant.in_memory_instance_managed_grpc_env(),
            GraphQLContextVariant.sqlite_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.sqlite_read_only_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.sqlite_with_default_run_launcher_deployed_grpc_env(),
            GraphQLContextVariant.sqlite_with_queued_run_coordinator_managed_grpc_env(),
            GraphQLContextVariant.postgres_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.postgres_with_default_run_launcher_deployed_grpc_env(),
            GraphQLContextVariant.non_launchable_in_memory_instance_multi_location(),
            GraphQLContextVariant.non_launchable_in_memory_instance_managed_grpc_env(),
            GraphQLContextVariant.non_launchable_in_memory_instance_lazy_repository(),
            GraphQLContextVariant.non_launchable_sqlite_instance_multi_location(),
            GraphQLContextVariant.non_launchable_sqlite_instance_managed_grpc_env(),
            GraphQLContextVariant.non_launchable_sqlite_instance_deployed_grpc_env(),
            GraphQLContextVariant.non_launchable_sqlite_instance_lazy_repository(),
            GraphQLContextVariant.non_launchable_postgres_instance_multi_location(),
            GraphQLContextVariant.non_launchable_postgres_instance_managed_grpc_env(),
            GraphQLContextVariant.non_launchable_postgres_instance_lazy_repository(),
            GraphQLContextVariant.consolidated_sqlite_instance_managed_grpc_env(),
        ]

    @staticmethod
    def all_executing_variants():
        return [
            GraphQLContextVariant.sqlite_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.sqlite_with_default_run_launcher_deployed_grpc_env(),
            GraphQLContextVariant.postgres_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.postgres_with_default_run_launcher_deployed_grpc_env(),
        ]

    @staticmethod
    def all_readonly_variants():
        """
        Return all read only variants. If you try to run any mutation these will error
        """
        return _variants_with_mark(GraphQLContextVariant.all_variants(), pytest.mark.read_only)

    @staticmethod
    def all_non_launchable_variants():
        """
        Return all non_launchable variants. If you try to start or launch these will error
        """
        return _variants_with_mark(GraphQLContextVariant.all_variants(), pytest.mark.non_launchable)

    @staticmethod
    def all_multi_location_variants():
        return _variants_with_mark(GraphQLContextVariant.all_variants(), pytest.mark.multi_location)


def _variants_with_mark(variants, mark):
    def _yield_all():
        for variant in variants:
            if mark in variant.marks:
                yield variant

    return list(_yield_all())


def _variants_without_marks(variants, marks):
    def _yield_all():
        for variant in variants:
            if all(mark not in variant.marks for mark in marks):
                yield variant

    return list(_yield_all())


@contextmanager
def manage_graphql_context(context_variant):
    with context_variant.instance_mgr() as instance:
        with context_variant.environment_mgr(
            instance, context_variant.read_only
        ) as workspace_process_context:
            yield workspace_process_context.create_request_context()


class _GraphQLContextTestSuite(ABC):
    @abstractmethod
    def yield_graphql_context(self, request):
        pass

    @contextmanager
    def graphql_context_for_request(self, request):
        check.param_invariant(
            isinstance(request.param, GraphQLContextVariant),
            "request",
            "params in fixture must be List[GraphQLContextVariant]",
        )
        with manage_graphql_context(request.param) as graphql_context:
            yield graphql_context


def graphql_context_variants_fixture(context_variants):
    check.list_param(context_variants, "context_variants", of_type=GraphQLContextVariant)

    def _wrap(fn):
        return pytest.fixture(
            name="graphql_context",
            params=[
                pytest.param(
                    context_variant,
                    id=context_variant.test_id,
                    marks=context_variant.marks + [Marks.graphql_context_test_suite],
                )
                for context_variant in context_variants
            ],
        )(fn)

    return _wrap


def make_graphql_context_test_suite(context_variants):
    """
        Arguments:

        runs (List[GraphQLContextVariant]): List of runs to run per test in this class.

        This is the base class factory for test suites in the dagster-graphql test.

        The goal of this suite is to make it straightforward to run tests
        against multiple graphql_contexts, have a coherent lifecycle for those
        contexts.

        GraphQLContextVariant has a number of static methods to provide common run configurations
        as well as common groups of run configuration

        One can also make bespoke GraphQLContextVariants which specific implementations
        of DagsterInstance, RepositoryLocation, and so forth. See that class
        for more details.

    Example:

    class TestAThing(
        make_graphql_context_test_suite(
            context_variants=[GraphQLContextVariant.in_memory_in_process_start()]
        )
    ):
        def test_graphql_context_exists(self, graphql_context):
            assert graphql_context
    """
    check.list_param(context_variants, "context_variants", of_type=GraphQLContextVariant)

    class _SpecificTestSuiteBase(_GraphQLContextTestSuite):
        @graphql_context_variants_fixture(context_variants=context_variants)
        def yield_graphql_context(self, request):
            with self.graphql_context_for_request(request) as graphql_context:
                yield graphql_context

        @pytest.fixture(name="graphql_client")
        def yield_graphql_client(self, graphql_context):
            class MockedGraphQLClient:
                def execute(self, gql_query, variable_values=None):
                    return execute_dagster_graphql(
                        graphql_context,
                        gql_query,
                        variable_values,
                    ).data

            with patch("dagster_graphql.client.client.Client") as mock_client:
                mock_client.return_value = MockedGraphQLClient()
                yield DagsterGraphQLClient("localhost")

    return _SpecificTestSuiteBase


ReadonlyGraphQLContextTestMatrix = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_readonly_variants()
)


NonLaunchableGraphQLContextTestMatrix = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_non_launchable_variants()
)

ExecutingGraphQLContextTestMatrix = make_graphql_context_test_suite(
    context_variants=GraphQLContextVariant.all_executing_variants()
)



from dagster import check
from dagster.core.definitions.run_request import InstigatorType
from dagster.core.host_representation import InstigationSelector
from dagster.core.scheduler.instigation import InstigatorStatus

from .utils import capture_error


@capture_error
def get_unloadable_job_states_or_error(graphene_info, job_type=None):
    from ..schema.instigation import GrapheneInstigationState, GrapheneInstigationStates

    check.opt_inst_param(job_type, "job_type", InstigatorType)
    job_states = graphene_info.context.instance.all_stored_job_state(job_type=job_type)
    external_jobs = [
        job
        for repository_location in graphene_info.context.repository_locations
        for repository in repository_location.get_repositories().values()
        for job in repository.get_external_schedules() + repository.get_external_sensors()
    ]

    job_origin_ids = {job.get_external_origin_id() for job in external_jobs}

    unloadable_states = [
        job_state
        for job_state in job_states
        if job_state.job_origin_id not in job_origin_ids
        and job_state.status == InstigatorStatus.RUNNING
    ]

    return GrapheneInstigationStates(
        results=[GrapheneInstigationState(job_state=job_state) for job_state in unloadable_states]
    )


@capture_error
def get_job_state_or_error(graphene_info, selector):
    from ..schema.instigation import GrapheneInstigationState

    check.inst_param(selector, "selector", InstigationSelector)
    location = graphene_info.context.get_repository_location(selector.location_name)
    repository = location.get_repository(selector.repository_name)

    if repository.has_external_sensor(selector.name):
        external_sensor = repository.get_external_sensor(selector.name)
        job_state = graphene_info.context.instance.get_job_state(
            external_sensor.get_external_origin_id()
        )
        if not job_state:
            job_state = external_sensor.get_default_instigation_state()
    elif repository.has_external_schedule(selector.name):
        external_schedule = repository.get_external_schedule(selector.name)
        job_state = graphene_info.context.instance.get_job_state(
            external_schedule.get_external_origin_id()
        )
        if not job_state:
            job_state = external_schedule.get_default_instigation_state()
    else:
        check.failed(f"Could not find a definition for {selector.name}")

    return GrapheneInstigationState(job_state=job_state)



import pytest
from dagster import DagsterInvalidDefinitionError
from dagster_graphql import DagsterGraphQLClientError, InvalidOutputErrorInfo

from .conftest import MockClient, python_client_test_suite

EXPECTED_RUN_ID = "foo"

launch_pipeline_success_response = {
    "launchPipelineExecution": {
        "__typename": "LaunchRunSuccess",
        "run": {"runId": EXPECTED_RUN_ID},
    }
}


@python_client_test_suite
def test_success(mock_client: MockClient):
    mock_client.mock_gql_client.execute.return_value = launch_pipeline_success_response
    actual_run_id = mock_client.python_client.submit_pipeline_execution(
        "bar",
        repository_location_name="baz",
        repository_name="quux",
        run_config={},
        mode="default",
    )
    assert actual_run_id == EXPECTED_RUN_ID


@python_client_test_suite
def test_job_success(mock_client: MockClient):
    mock_client.mock_gql_client.execute.return_value = launch_pipeline_success_response
    actual_run_id = mock_client.python_client.submit_job_execution(
        "bar",
        repository_location_name="baz",
        repository_name="quux",
    )
    assert actual_run_id == EXPECTED_RUN_ID


@python_client_test_suite
def test_preset_success(mock_client: MockClient):
    mock_client.mock_gql_client.execute.return_value = launch_pipeline_success_response
    actual_run_id = mock_client.python_client.submit_pipeline_execution(
        "bar", repository_location_name="baz", repository_name="quux", preset="cool_preset"
    )
    assert actual_run_id == EXPECTED_RUN_ID


@python_client_test_suite
def test_tags_success(mock_client: MockClient):
    mock_client.mock_gql_client.execute.return_value = launch_pipeline_success_response
    actual_run_id = mock_client.python_client.submit_pipeline_execution(
        "bar",
        repository_location_name="baz",
        repository_name="quuz",
        run_config={},
        mode="default",
        tags={"my_tag": "a", "my_other_tag": "b"},
    )
    assert actual_run_id == EXPECTED_RUN_ID


@python_client_test_suite
def test_job_tags_success(mock_client: MockClient):
    mock_client.mock_gql_client.execute.return_value = launch_pipeline_success_response
    actual_run_id = mock_client.python_client.submit_job_execution(
        "bar",
        repository_location_name="baz",
        repository_name="quuz",
        tags={"my_tag": "a", "my_other_tag": "b"},
    )
    assert actual_run_id == EXPECTED_RUN_ID


@python_client_test_suite
def test_complex_tags_success(mock_client: MockClient):
    response = {
        "launchPipelineExecution": {
            "__typename": "LaunchRunSuccess",
            "run": {"runId": EXPECTED_RUN_ID},
        }
    }
    mock_client.mock_gql_client.execute.return_value = response
    actual_run_id = mock_client.python_client.submit_pipeline_execution(
        "bar",
        repository_location_name="baz",
        repository_name="quuz",
        run_config={},
        mode="default",
        tags={"my_tag": {"I'm": {"a JSON-encodable": "thing"}}},
    )
    assert actual_run_id == EXPECTED_RUN_ID

    actual_run_id = mock_client.python_client.submit_job_execution(
        "bar",
        repository_location_name="baz",
        repository_name="quuz",
        run_config={},
        tags={"my_tag": {"I'm": {"a JSON-encodable": "thing"}}},
    )
    assert actual_run_id == EXPECTED_RUN_ID


@python_client_test_suite
def test_invalid_tags_failure(mock_client: MockClient):
    class SomeWeirdObject:
        pass

    with pytest.raises(DagsterInvalidDefinitionError):
        mock_client.python_client.submit_pipeline_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quuz",
            run_config={},
            mode="default",
            tags={"my_invalid_tag": SomeWeirdObject()},
        )

    with pytest.raises(DagsterInvalidDefinitionError):
        mock_client.python_client.submit_job_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quuz",
            run_config={},
            tags={"my_invalid_tag": SomeWeirdObject()},
        )


@python_client_test_suite
def test_no_location_or_repo_provided_success(mock_client: MockClient):
    repo_loc_name, repo_name, pipeline_name = "bar", "baz", "quux"
    other_repo_name, other_pipeline_name = "other repo", "my_pipeline"
    get_locations_and_names_response = {
        "repositoriesOrError": {
            "__typename": "RepositoryConnection",
            "nodes": [
                {
                    "name": repo_name,
                    "location": {"name": repo_loc_name},
                    "pipelines": [{"name": pipeline_name}, {"name": other_pipeline_name}],
                },
                {
                    "name": other_repo_name,
                    "location": {"name": repo_loc_name},
                    "pipelines": [{"name": "fun pipeline"}, {"name": other_pipeline_name}],
                },
            ],
        }
    }
    submit_execution_response = {
        "launchPipelineExecution": {
            "__typename": "LaunchRunSuccess",
            "run": {"runId": EXPECTED_RUN_ID},
        }
    }
    mock_client.mock_gql_client.execute.side_effect = [
        get_locations_and_names_response,
        submit_execution_response,
    ]

    actual_run_id = mock_client.python_client.submit_pipeline_execution(
        pipeline_name, run_config={}, mode="default"
    )
    assert actual_run_id == EXPECTED_RUN_ID

    mock_client.mock_gql_client.execute.side_effect = [
        get_locations_and_names_response,
        submit_execution_response,
    ]

    actual_run_id = mock_client.python_client.submit_job_execution(pipeline_name, run_config={})
    assert actual_run_id == EXPECTED_RUN_ID


def no_location_or_repo_provided_duplicate_pipeline_mock_config(mock_client: MockClient):
    repo_loc_name, repo_name, pipeline_name = "bar", "baz", "quux"
    other_repo_name = "other repo"
    get_locations_and_names_response = {
        "repositoriesOrError": {
            "__typename": "RepositoryConnection",
            "nodes": [
                {
                    "name": repo_name,
                    "location": {"name": repo_loc_name},
                    "pipelines": [{"name": pipeline_name}],
                },
                {
                    "name": other_repo_name,
                    "location": {"name": repo_loc_name},
                    "pipelines": [{"name": pipeline_name}],
                },
            ],
        }
    }
    submit_execution_response = {
        "launchPipelineExecution": {
            "__typename": "LaunchRunSuccess",
            "run": {"runId": EXPECTED_RUN_ID},
        }
    }
    mock_client.mock_gql_client.execute.side_effect = [
        get_locations_and_names_response,
        submit_execution_response,
    ]

    return pipeline_name


@python_client_test_suite
def test_no_location_or_repo_provided_duplicate_pipeline_failure(mock_client: MockClient):
    pipeline_name = no_location_or_repo_provided_duplicate_pipeline_mock_config(mock_client)

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_pipeline_execution(
            pipeline_name, run_config={}, mode="default"
        )

    assert exc_info.value.args[0].find(f"multiple pipelines with the name {pipeline_name}") != -1


@python_client_test_suite
def test_no_location_or_repo_provided_duplicate_job_failure(mock_client: MockClient):
    job_name = no_location_or_repo_provided_duplicate_pipeline_mock_config(mock_client)

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_job_execution(job_name, run_config={})

    assert exc_info.value.args[0].find(f"multiple jobs with the name {job_name}") != -1


def no_location_or_repo_provided_mock_config(mock_client):
    repo_loc_name, repo_name, pipeline_name = "bar", "baz", "quux"
    get_locations_and_names_response = {
        "repositoriesOrError": {
            "__typename": "RepositoryConnection",
            "nodes": [
                {
                    "name": repo_name,
                    "location": {"name": repo_loc_name},
                    "pipelines": [{"name": pipeline_name}],
                }
            ],
        }
    }
    submit_execution_response = {
        "launchPipelineExecution": {
            "__typename": "LaunchRunSuccess",
            "run": {"runId": EXPECTED_RUN_ID},
        }
    }
    mock_client.mock_gql_client.execute.side_effect = [
        get_locations_and_names_response,
        submit_execution_response,
    ]


@python_client_test_suite
def test_no_location_or_repo_provided_no_pipeline_failure(mock_client: MockClient):
    no_location_or_repo_provided_mock_config(mock_client)

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_pipeline_execution("123", run_config={}, mode="default")

    assert exc_info.value.args[0] == "PipelineNotFoundError"


@python_client_test_suite
def test_no_location_or_repo_provided_no_job_failure(mock_client: MockClient):
    no_location_or_repo_provided_mock_config(mock_client)

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_job_execution("123", run_config={})

    assert exc_info.value.args[0] == "JobNotFoundError"


@python_client_test_suite
def test_failure_with_invalid_step_error(mock_client: MockClient):
    error_type, invalid_step_key = "InvalidStepError", "1234"
    response = {
        "launchPipelineExecution": {
            "__typename": error_type,
            "invalidStepKey": invalid_step_key,
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_pipeline_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quux",
            run_config={},
            mode="default",
        )
    exc_args = exc_info.value.args

    assert exc_args[0] == error_type
    assert exc_args[1] == invalid_step_key


@python_client_test_suite
def test_failure_with_invalid_output_error(mock_client: MockClient):
    error_type, step_key, invalid_output_name = "InvalidOutputError", "1234", "some output"
    response = {
        "launchPipelineExecution": {
            "__typename": error_type,
            "stepKey": step_key,
            "invalidOutputName": invalid_output_name,
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_pipeline_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quux",
            run_config={},
            mode="default",
        )

    assert exc_info.value.args == (error_type,)
    assert exc_info.value.body == InvalidOutputErrorInfo(
        step_key=step_key, invalid_output_name=invalid_output_name
    )


@python_client_test_suite
def test_failure_with_pipeline_config_invalid(mock_client: MockClient):
    error_type = "RunConfigValidationInvalid"
    errors = [
        {
            "__typename": "some_error",
            "message": "AWS warehouse got hit by a meteor",
            "path": [],
            "reason": "Network failure",
        }
    ]
    response = {"launchPipelineExecution": {"__typename": error_type, "errors": errors}}
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_pipeline_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quux",
            run_config={},
            mode="default",
        )
    exc_args = exc_info.value.args

    assert exc_args[0] == error_type
    assert exc_args[1] == errors


@python_client_test_suite
def test_failure_with_python_error(mock_client: MockClient):
    error_type, message = "PythonError", "some catastrophic error"
    response = {
        "launchPipelineExecution": {
            "__typename": error_type,
            "message": message,
        }
    }
    mock_client.mock_gql_client.execute.return_value = response

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_pipeline_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quux",
            run_config={},
            mode="default",
        )
    exc_args = exc_info.value.args

    assert exc_args[0] == error_type
    assert exc_args[1] == message

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_job_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quux",
            run_config={},
        )
    exc_args = exc_info.value.args

    assert exc_args[0] == error_type
    assert exc_args[1] == message


def failure_with_pipeline_run_conflict_mock_config(mock_client: MockClient):
    error_type, message = "RunConflict", "some conflict"
    response = {
        "launchPipelineExecution": {
            "__typename": error_type,
            "message": message,
        }
    }
    mock_client.mock_gql_client.execute.return_value = response


@python_client_test_suite
def test_failure_with_pipeline_run_conflict(mock_client: MockClient):
    failure_with_pipeline_run_conflict_mock_config(mock_client)

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_pipeline_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quux",
            run_config={},
            mode="default",
        )
    exc_args = exc_info.value.args

    assert exc_args[0] == "RunConflict"
    assert exc_args[1] == "some conflict"


@python_client_test_suite
def test_failure_with_job_run_conflict(mock_client: MockClient):
    failure_with_pipeline_run_conflict_mock_config(mock_client)

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_job_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quux",
            run_config={},
        )
    exc_args = exc_info.value.args

    assert exc_args[0] == "RunConflict"
    assert exc_args[1] == "some conflict"


@python_client_test_suite
def test_failure_with_query_error(mock_client: MockClient):
    mock_client.mock_gql_client.side_effect = Exception("foo")

    with pytest.raises(DagsterGraphQLClientError) as exc_info:
        mock_client.python_client.submit_pipeline_execution(
            "bar",
            repository_location_name="baz",
            repository_name="quux",
            run_config={},
            mode="default",
        )

    assert exc_info.value.args[0].endswith("failed GraphQL validation")



import graphene
from dagster import check
from dagster.core.storage.compute_log_manager import ComputeIOType

from ...implementation.execution import get_compute_log_observable, get_pipeline_run_observable
from ..external import GrapheneLocationStateChangeSubscription, get_location_state_change_observable
from ..logs.compute_logs import GrapheneComputeIOType, GrapheneComputeLogFile
from ..paging import GrapheneCursor
from ..pipelines.subscription import GraphenePipelineRunLogsSubscriptionPayload


class GrapheneDagitSubscription(graphene.ObjectType):
    class Meta:
        name = "DagitSubscription"

    pipelineRunLogs = graphene.Field(
        graphene.NonNull(GraphenePipelineRunLogsSubscriptionPayload),
        runId=graphene.Argument(graphene.NonNull(graphene.ID)),
        after=graphene.Argument(GrapheneCursor),
    )

    computeLogs = graphene.Field(
        graphene.NonNull(GrapheneComputeLogFile),
        runId=graphene.Argument(graphene.NonNull(graphene.ID)),
        stepKey=graphene.Argument(graphene.NonNull(graphene.String)),
        ioType=graphene.Argument(graphene.NonNull(GrapheneComputeIOType)),
        cursor=graphene.Argument(graphene.String),
    )

    locationStateChangeEvents = graphene.Field(
        graphene.NonNull(GrapheneLocationStateChangeSubscription)
    )

    def resolve_pipelineRunLogs(self, graphene_info, runId, after=None):
        return get_pipeline_run_observable(graphene_info, runId, after)

    def resolve_computeLogs(self, graphene_info, runId, stepKey, ioType, cursor=None):
        check.str_param(ioType, "ioType")  # need to resolve to enum
        return get_compute_log_observable(
            graphene_info, runId, stepKey, ComputeIOType(ioType), cursor
        )

    def resolve_locationStateChangeEvents(self, graphene_info):
        return get_location_state_change_observable(graphene_info)



import os
import time

from dagster import AssetKey
from dagster.utils import safe_tempfile_path
from dagster_graphql.client.query import LAUNCH_PIPELINE_EXECUTION_MUTATION
from dagster_graphql.test.utils import (
    execute_dagster_graphql,
    infer_pipeline_selector,
    infer_repository_selector,
)

from .graphql_context_test_suite import GraphQLContextVariant, make_graphql_context_test_suite

GET_ASSET_KEY_QUERY = """
    query AssetKeyQuery {
        assetsOrError {
            __typename
            ...on AssetConnection {
                nodes {
                    key {
                        path
                    }
                }
            }
        }
    }
"""

GET_ASSET_MATERIALIZATION = """
    query AssetQuery($assetKey: AssetKeyInput!) {
        assetOrError(assetKey: $assetKey) {
            ... on Asset {
                assetMaterializations(limit: 1) {
                    materializationEvent {
                        materialization {
                            label
                        }
                        assetLineage {
                            assetKey {
                                path
                            }
                            partitions
                        }
                    }
                }
            }
            ... on AssetNotFoundError {
                __typename
            }
        }
    }
"""

GET_ASSET_MATERIALIZATION_WITH_PARTITION = """
    query AssetQuery($assetKey: AssetKeyInput!) {
        assetOrError(assetKey: $assetKey) {
            ... on Asset {
                assetMaterializations(limit: 1) {
                    partition
                    materializationEvent {
                        materialization {
                            label
                        }
                    }
                }
            }
        }
    }
"""


WIPE_ASSETS = """
    mutation AssetKeyWipe($assetKeys: [AssetKeyInput!]!) {
        wipeAssets(assetKeys: $assetKeys) {
            __typename
        }
    }
"""

GET_ASSET_MATERIALIZATION_TIMESTAMP = """
    query AssetQuery($assetKey: AssetKeyInput!, $asOf: String) {
        assetOrError(assetKey: $assetKey) {
            ... on Asset {
                assetMaterializations(beforeTimestampMillis: $asOf) {
                    materializationEvent {
                        timestamp
                    }
                }
            }
        }
    }
"""

GET_ASSET_IN_PROGRESS_RUNS = """
    query AssetGraphQuery($repositorySelector: RepositorySelector!) {
        repositoryOrError(repositorySelector: $repositorySelector) {
            ... on Repository {
                assetNodes {
                    opName
                    description
                    jobs {
                        id
                        name
                    }
                }
                inProgressRunsByStep {
                    stepKey
                    unstartedRuns {
                        runId
                    }
                    inProgressRuns {
                        runId
                    }
                }
            }
        }
    }
"""

GET_ASSET_NODES_FROM_KEYS = """
    query AssetNodeQuery($pipelineSelector: PipelineSelector!, $assetKeys: [AssetKeyInput!]) {
        pipelineOrError(params: $pipelineSelector) {
            ... on Pipeline {
                id
                assetNodes(assetKeys: $assetKeys) {
                    id
                }
            }
        }
    }
"""


GET_ASSET_PARTITIONS_FROM_KEYS = """
    query AssetNodeQuery($pipelineSelector: PipelineSelector!) {
        pipelineOrError(params: $pipelineSelector) {
            ... on Pipeline {
                id
                assetNodes {
                    id
                    partitionKeys
                }
            }
        }
    }
"""

GET_LATEST_MATERIALIZATION_PER_PARTITION = """
    query AssetNodeQuery($pipelineSelector: PipelineSelector!, $partitions: [String!]) {
        pipelineOrError(params: $pipelineSelector) {
            ... on Pipeline {
                id
                assetNodes {
                    id
                    partitionKeys
                    latestMaterializationByPartition(partitions: $partitions) {
                        partition
                        materializationEvent {
                            stepStats {
                                startTime
                            }
                        }
                    }
                }
            }
        }
    }
"""


def _create_run(graphql_context, pipeline_name, mode="default"):
    selector = infer_pipeline_selector(graphql_context, pipeline_name)
    result = execute_dagster_graphql(
        graphql_context,
        LAUNCH_PIPELINE_EXECUTION_MUTATION,
        variables={
            "executionParams": {
                "selector": selector,
                "mode": mode,
            }
        },
    )
    assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"
    graphql_context.instance.run_launcher.join()
    return result.data["launchPipelineExecution"]["run"]["runId"]


class TestAssetAwareEventLog(
    make_graphql_context_test_suite(
        context_variants=[
            GraphQLContextVariant.consolidated_sqlite_instance_managed_grpc_env(),
            GraphQLContextVariant.sqlite_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.postgres_with_default_run_launcher_managed_grpc_env(),
        ]
    )
):
    def test_all_asset_keys(self, graphql_context, snapshot):
        _create_run(graphql_context, "multi_asset_pipeline")
        result = execute_dagster_graphql(graphql_context, GET_ASSET_KEY_QUERY)
        assert result.data
        assert result.data["assetsOrError"]
        assert result.data["assetsOrError"]["nodes"]

        # sort by materialization asset key to keep list order is consistent for snapshot
        result.data["assetsOrError"]["nodes"].sort(key=lambda e: e["key"]["path"][0])

        snapshot.assert_match(result.data)

    def test_get_asset_key_materialization(self, graphql_context, snapshot):
        _create_run(graphql_context, "single_asset_pipeline")
        result = execute_dagster_graphql(
            graphql_context, GET_ASSET_MATERIALIZATION, variables={"assetKey": {"path": ["a"]}}
        )
        assert result.data
        snapshot.assert_match(result.data)

    def test_get_asset_key_not_found(self, graphql_context, snapshot):
        _create_run(graphql_context, "single_asset_pipeline")

        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_MATERIALIZATION,
            variables={"assetKey": {"path": ["bogus", "asset"]}},
        )
        assert result.data
        snapshot.assert_match(result.data)

    def test_get_partitioned_asset_key_materialization(self, graphql_context, snapshot):
        _create_run(graphql_context, "partitioned_asset_pipeline")

        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_MATERIALIZATION_WITH_PARTITION,
            variables={"assetKey": {"path": ["a"]}},
        )
        assert result.data
        snapshot.assert_match(result.data)

    def test_get_asset_key_lineage(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, "asset_lineage_pipeline")
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={"executionParams": {"selector": selector, "mode": "default"}},
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        graphql_context.instance.run_launcher.join()

        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_MATERIALIZATION,
            variables={"assetKey": {"path": ["b"]}},
        )
        assert result.data
        snapshot.assert_match(result.data)

    def test_get_partitioned_asset_key_lineage(self, graphql_context, snapshot):
        selector = infer_pipeline_selector(graphql_context, "partitioned_asset_lineage_pipeline")
        result = execute_dagster_graphql(
            graphql_context,
            LAUNCH_PIPELINE_EXECUTION_MUTATION,
            variables={"executionParams": {"selector": selector, "mode": "default"}},
        )
        assert result.data["launchPipelineExecution"]["__typename"] == "LaunchRunSuccess"

        graphql_context.instance.run_launcher.join()

        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_MATERIALIZATION,
            variables={"assetKey": {"path": ["b"]}},
        )
        assert result.data
        snapshot.assert_match(result.data)

    def test_asset_wipe(self, graphql_context):
        _create_run(graphql_context, "single_asset_pipeline")
        _create_run(graphql_context, "multi_asset_pipeline")

        asset_keys = graphql_context.instance.all_asset_keys()
        assert AssetKey("a") in asset_keys

        result = execute_dagster_graphql(
            graphql_context, WIPE_ASSETS, variables={"assetKeys": [{"path": ["a"]}]}
        )

        assert result.data
        assert result.data["wipeAssets"]
        assert result.data["wipeAssets"]["__typename"] == "AssetWipeSuccess"

        asset_keys = graphql_context.instance.all_asset_keys()
        assert AssetKey("a") not in asset_keys

    def test_asset_asof_timestamp(self, graphql_context):
        _create_run(graphql_context, "asset_tag_pipeline")
        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_MATERIALIZATION_TIMESTAMP,
            variables={"assetKey": {"path": ["a"]}},
        )
        assert result.data
        assert result.data["assetOrError"]
        materializations = result.data["assetOrError"]["assetMaterializations"]
        assert len(materializations) == 1
        first_timestamp = int(materializations[0]["materializationEvent"]["timestamp"])

        as_of_timestamp = first_timestamp + 1

        time.sleep(1.1)
        _create_run(graphql_context, "asset_tag_pipeline")
        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_MATERIALIZATION_TIMESTAMP,
            variables={"assetKey": {"path": ["a"]}},
        )
        assert result.data
        assert result.data["assetOrError"]
        materializations = result.data["assetOrError"]["assetMaterializations"]
        assert len(materializations) == 2
        second_timestamp = int(materializations[0]["materializationEvent"]["timestamp"])

        assert second_timestamp > as_of_timestamp

        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_MATERIALIZATION_TIMESTAMP,
            variables={"assetKey": {"path": ["a"]}, "asOf": as_of_timestamp},
        )
        assert result.data
        assert result.data["assetOrError"]
        materializations = result.data["assetOrError"]["assetMaterializations"]
        assert len(materializations) == 1
        assert first_timestamp == int(materializations[0]["materializationEvent"]["timestamp"])

    def test_asset_node_in_pipeline(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "two_assets_job")
        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_NODES_FROM_KEYS,
            variables={"pipelineSelector": selector, "assetKeys": [{"path": ["asset_one"]}]},
        )

        assert result.data
        assert result.data["pipelineOrError"]
        assert result.data["pipelineOrError"]["assetNodes"]

        assert len(result.data["pipelineOrError"]["assetNodes"]) == 1
        asset_node = result.data["pipelineOrError"]["assetNodes"][0]
        assert asset_node["id"] == '["asset_one"]'

        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_NODES_FROM_KEYS,
            variables={"pipelineSelector": selector},
        )

        assert result.data
        assert result.data["pipelineOrError"]
        assert result.data["pipelineOrError"]["assetNodes"]

        assert len(result.data["pipelineOrError"]["assetNodes"]) == 2
        asset_node = result.data["pipelineOrError"]["assetNodes"][0]
        assert asset_node["id"] == '["asset_one"]'

    def test_asset_partitions_in_pipeline(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "two_assets_job")
        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_PARTITIONS_FROM_KEYS,
            variables={"pipelineSelector": selector},
        )

        assert result.data
        assert result.data["pipelineOrError"]
        assert result.data["pipelineOrError"]["assetNodes"]
        assert len(result.data["pipelineOrError"]["assetNodes"]) == 2
        asset_node = result.data["pipelineOrError"]["assetNodes"][0]
        assert asset_node["partitionKeys"] == []

        selector = infer_pipeline_selector(graphql_context, "static_partitioned_assets_job")
        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_PARTITIONS_FROM_KEYS,
            variables={"pipelineSelector": selector},
        )

        assert result.data
        assert result.data["pipelineOrError"]
        assert result.data["pipelineOrError"]["assetNodes"]
        assert len(result.data["pipelineOrError"]["assetNodes"]) == 2
        asset_node = result.data["pipelineOrError"]["assetNodes"][0]
        assert asset_node["partitionKeys"] and asset_node["partitionKeys"] == [
            "a",
            "b",
            "c",
            "d",
        ]
        asset_node = result.data["pipelineOrError"]["assetNodes"][1]
        assert asset_node["partitionKeys"] and asset_node["partitionKeys"] == [
            "a",
            "b",
            "c",
            "d",
        ]

        selector = infer_pipeline_selector(graphql_context, "time_partitioned_assets_job")
        result = execute_dagster_graphql(
            graphql_context,
            GET_ASSET_PARTITIONS_FROM_KEYS,
            variables={"pipelineSelector": selector},
        )

        assert result.data
        assert result.data["pipelineOrError"]
        assert result.data["pipelineOrError"]["assetNodes"]
        assert len(result.data["pipelineOrError"]["assetNodes"]) == 2
        asset_node = result.data["pipelineOrError"]["assetNodes"][0]

        # test partition starts at "2021-05-05-01:00". Should be > 100 partition keys
        # since partition is hourly
        assert asset_node["partitionKeys"] and len(asset_node["partitionKeys"]) > 100
        assert asset_node["partitionKeys"][0] == "2021-05-05-01:00"
        assert asset_node["partitionKeys"][1] == "2021-05-05-02:00"

    def test_latest_materialization_per_partition(self, graphql_context):
        _create_run(graphql_context, "partition_materialization_job")

        selector = infer_pipeline_selector(graphql_context, "partition_materialization_job")
        result = execute_dagster_graphql(
            graphql_context,
            GET_LATEST_MATERIALIZATION_PER_PARTITION,
            variables={"pipelineSelector": selector, "partitions": ["a"]},
        )

        assert result.data
        assert result.data["pipelineOrError"]
        assert result.data["pipelineOrError"]["assetNodes"]
        asset_node = result.data["pipelineOrError"]["assetNodes"][0]
        assert len(asset_node["latestMaterializationByPartition"]) == 1
        assert asset_node["latestMaterializationByPartition"][0] == None

        result = execute_dagster_graphql(
            graphql_context,
            GET_LATEST_MATERIALIZATION_PER_PARTITION,
            variables={"pipelineSelector": selector, "partitions": ["c"]},
        )

        assert result.data
        assert result.data["pipelineOrError"]
        assert result.data["pipelineOrError"]["assetNodes"]
        asset_node = result.data["pipelineOrError"]["assetNodes"][0]
        assert len(asset_node["latestMaterializationByPartition"]) == 1
        materialization = asset_node["latestMaterializationByPartition"][0]
        start_time = materialization["materializationEvent"]["stepStats"]["startTime"]
        assert materialization["partition"] == "c"

        _create_run(graphql_context, "partition_materialization_job")
        result = execute_dagster_graphql(
            graphql_context,
            GET_LATEST_MATERIALIZATION_PER_PARTITION,
            variables={"pipelineSelector": selector, "partitions": ["c", "a"]},
        )
        assert (
            result.data
            and result.data["pipelineOrError"]
            and result.data["pipelineOrError"]["assetNodes"]
        )
        asset_node = result.data["pipelineOrError"]["assetNodes"][0]
        assert len(asset_node["latestMaterializationByPartition"]) == 2
        materialization = asset_node["latestMaterializationByPartition"][0]
        new_start_time = materialization["materializationEvent"]["stepStats"]["startTime"]
        assert new_start_time > start_time

        assert asset_node["latestMaterializationByPartition"][1] == None


class TestPersistentInstanceAssetInProgress(
    make_graphql_context_test_suite(
        context_variants=[
            GraphQLContextVariant.sqlite_with_default_run_launcher_managed_grpc_env(),
            GraphQLContextVariant.postgres_with_default_run_launcher_managed_grpc_env(),
        ]
    )
):
    def test_asset_in_progress(self, graphql_context):
        selector = infer_pipeline_selector(graphql_context, "hanging_job")
        run_id = "foo"

        with safe_tempfile_path() as path:
            result = execute_dagster_graphql(
                graphql_context,
                LAUNCH_PIPELINE_EXECUTION_MUTATION,
                variables={
                    "executionParams": {
                        "selector": selector,
                        "mode": "default",
                        "runConfigData": {
                            "resources": {"hanging_asset_resource": {"config": {"file": path}}}
                        },
                        "executionMetadata": {"runId": run_id},
                    }
                },
            )

            assert not result.errors
            assert result.data

            # ensure the execution has happened
            while not os.path.exists(path):
                time.sleep(0.1)

            result = execute_dagster_graphql(
                graphql_context,
                GET_ASSET_IN_PROGRESS_RUNS,
                variables={"repositorySelector": infer_repository_selector(graphql_context)},
            )
            graphql_context.instance.run_launcher.terminate(run_id)

            assert result.data
            assert result.data["repositoryOrError"]
            assert result.data["repositoryOrError"]["inProgressRunsByStep"]

            in_progress_runs_by_step = result.data["repositoryOrError"]["inProgressRunsByStep"]

            assert len(in_progress_runs_by_step) == 2

            hanging_asset_status = in_progress_runs_by_step[0]
            never_runs_asset_status = in_progress_runs_by_step[1]
            # graphql endpoint returns unordered list of steps
            # swap if never_runs_asset_status is first in list
            if hanging_asset_status["stepKey"] != "hanging_asset":
                never_runs_asset_status, hanging_asset_status = (
                    hanging_asset_status,
                    never_runs_asset_status,
                )

            assert hanging_asset_status["stepKey"] == "hanging_asset"
            assert len(hanging_asset_status["inProgressRuns"]) == 1
            assert hanging_asset_status["inProgressRuns"][0]["runId"] == run_id
            assert len(hanging_asset_status["unstartedRuns"]) == 0

            assert never_runs_asset_status["stepKey"] == "never_runs_asset"
            assert len(never_runs_asset_status["inProgressRuns"]) == 0
            assert len(never_runs_asset_status["unstartedRuns"]) == 1
            assert never_runs_asset_status["unstartedRuns"][0]["runId"] == run_id



import time

from dagster.core.errors import DagsterUserCodeUnreachableError
from dagster_graphql import ShutdownRepositoryLocationStatus

from ..graphql.graphql_context_test_suite import (
    GraphQLContextVariant,
    make_graphql_context_test_suite,
)


class TestShutdownRepositoryLocation(
    make_graphql_context_test_suite(
        context_variants=[GraphQLContextVariant.non_launchable_sqlite_instance_deployed_grpc_env()]
    )
):
    def test_shutdown_repository_location(self, graphql_client, graphql_context):
        origin = next(iter(graphql_context.get_workspace_snapshot().values())).origin
        origin.create_client().heartbeat()

        result = graphql_client.shutdown_repository_location("test")

        assert result.status == ShutdownRepositoryLocationStatus.SUCCESS, result.message

        # Wait for client to be unavailable
        start_time = time.time()

        while time.time() - start_time < 15:
            try:
                origin.create_client().heartbeat()
            except DagsterUserCodeUnreachableError:
                # Shutdown succeeded
                return
            time.sleep(1)

        raise Exception("Timed out waiting for shutdown to take effect")

    def test_shutdown_repository_location_not_found(self, graphql_client):
        result = graphql_client.shutdown_repository_location("not_real")

        assert result.status == ShutdownRepositoryLocationStatus.FAILURE
        assert "Location not_real does not exist" in result.message



import graphene
from dagster import AssetKey, check
from dagster.core.events.log import EventLogEntry
from dagster.core.host_representation import ExternalRepository
from dagster.core.host_representation.external_data import (
    ExternalAssetNode,
    ExternalStaticPartitionsDefinitionData,
    ExternalTimeWindowPartitionsDefinitionData,
)

from . import external
from .asset_key import GrapheneAssetKey
from .errors import GrapheneAssetNotFoundError
from .pipelines.pipeline import GrapheneAssetMaterialization, GraphenePipeline
from .util import non_null_list


class GrapheneAssetDependency(graphene.ObjectType):
    class Meta:
        name = "AssetDependency"

    inputName = graphene.NonNull(graphene.String)
    asset = graphene.NonNull("dagster_graphql.schema.asset_graph.GrapheneAssetNode")

    def __init__(self, external_repository, input_name, asset_key):
        self._external_repository = check.inst_param(
            external_repository, "external_repository", ExternalRepository
        )
        self._asset_key = check.inst_param(asset_key, "asset_key", AssetKey)
        super().__init__(inputName=input_name)

    def resolve_asset(self, _graphene_info):
        return GrapheneAssetNode(
            self._external_repository,
            self._external_repository.get_external_asset_node(self._asset_key),
        )


class GrapheneAssetNode(graphene.ObjectType):
    id = graphene.NonNull(graphene.ID)
    assetKey = graphene.NonNull(GrapheneAssetKey)
    description = graphene.String()
    opName = graphene.String()
    jobs = non_null_list(GraphenePipeline)
    repository = graphene.NonNull(lambda: external.GrapheneRepository)
    dependencies = non_null_list(GrapheneAssetDependency)
    dependedBy = non_null_list(GrapheneAssetDependency)
    dependencyKeys = non_null_list(GrapheneAssetKey)
    dependedByKeys = non_null_list(GrapheneAssetKey)
    assetMaterializations = graphene.Field(
        non_null_list(GrapheneAssetMaterialization),
        partitions=graphene.List(graphene.String),
        beforeTimestampMillis=graphene.String(),
        limit=graphene.Int(),
    )
    partitionKeys = non_null_list(graphene.String)
    latestMaterializationByPartition = graphene.Field(
        graphene.NonNull(graphene.List(GrapheneAssetMaterialization)),
        partitions=graphene.List(graphene.String),
    )

    class Meta:
        name = "AssetNode"

    def __init__(
        self,
        external_repository,
        external_asset_node,
        latest_materialization=None,
        fetched_materialization=False,
    ):
        self._external_repository = check.inst_param(
            external_repository, "external_repository", ExternalRepository
        )
        self._external_asset_node = check.inst_param(
            external_asset_node, "external_asset_node", ExternalAssetNode
        )
        self._latest_materialization = check.opt_inst_param(
            latest_materialization, "latest_materialization", EventLogEntry
        )
        # we need a separate flag, because the asset might not have been materialized, so the None
        # value has significance
        self._fetched_materialization = check.bool_param(
            fetched_materialization, "fetched_materialization"
        )

        super().__init__(
            id=external_asset_node.asset_key.to_string(),
            assetKey=external_asset_node.asset_key,
            opName=external_asset_node.op_name,
            description=external_asset_node.op_description,
        )

    def resolve_repository(self, _graphene_info):
        loc = None
        for location in _graphene_info.context.repository_locations:
            if self._external_repository in location.get_repositories().values():
                loc = location
        return external.GrapheneRepository(self._external_repository, loc)

    def resolve_dependencies(self, _graphene_info):
        return [
            GrapheneAssetDependency(
                external_repository=self._external_repository,
                input_name=dep.input_name,
                asset_key=dep.upstream_asset_key,
            )
            for dep in self._external_asset_node.dependencies
        ]

    def resolve_dependedBy(self, _graphene_info):
        return [
            GrapheneAssetDependency(
                external_repository=self._external_repository,
                input_name=dep.input_name,
                asset_key=dep.downstream_asset_key,
            )
            for dep in self._external_asset_node.depended_by
        ]

    def resolve_dependencyKeys(self, _graphene_info):
        return [
            GrapheneAssetKey(path=dep.upstream_asset_key.path)
            for dep in self._external_asset_node.dependencies
        ]

    def resolve_dependencyByKeys(self, _graphene_info):
        return [
            GrapheneAssetKey(path=dep.downstream_asset_key.path)
            for dep in self._external_asset_node.depended_by
        ]

    def resolve_assetMaterializations(self, graphene_info, **kwargs):
        from ..implementation.fetch_assets import get_asset_events

        try:
            before_timestamp = (
                int(kwargs.get("beforeTimestampMillis")) / 1000.0
                if kwargs.get("beforeTimestampMillis")
                else None
            )
        except ValueError:
            before_timestamp = None

        limit = kwargs.get("limit")
        partitions = kwargs.get("partitions")
        if self._fetched_materialization and limit == 1 and not partitions and not before_timestamp:
            return (
                [GrapheneAssetMaterialization(event=self._latest_materialization)]
                if self._latest_materialization
                else []
            )

        return [
            GrapheneAssetMaterialization(event=event)
            for event in get_asset_events(
                graphene_info,
                self._external_asset_node.asset_key,
                partitions,
                before_timestamp=before_timestamp,
                limit=limit,
            )
        ]

    def resolve_jobs(self, _graphene_info):
        job_names = self._external_asset_node.job_names or []
        return [
            GraphenePipeline(self._external_repository.get_full_external_pipeline(job_name))
            for job_name in job_names
            if self._external_repository.has_external_pipeline(job_name)
        ]

    def resolve_partitionKeys(self, _graphene_info):
        # TODO: Add functionality for dynamic partitions definition
        partitions_def_data = self._external_asset_node.partitions_def_data
        if partitions_def_data:
            if isinstance(
                partitions_def_data, ExternalStaticPartitionsDefinitionData
            ) or isinstance(partitions_def_data, ExternalTimeWindowPartitionsDefinitionData):
                return [
                    partition.name
                    for partition in partitions_def_data.get_partitions_definition().get_partitions()
                ]
        return []

    def resolve_latestMaterializationByPartition(self, _graphene_info, **kwargs):
        from ..implementation.fetch_assets import get_asset_events

        get_partition = (
            lambda event: event.dagster_event.step_materialization_data.materialization.partition
        )

        partitions = kwargs.get("partitions")
        events_for_partitions = get_asset_events(
            _graphene_info,
            self._external_asset_node.asset_key,
            partitions,
        )

        latest_materialization_by_partition = {}
        for event in events_for_partitions:  # events are sorted in order of newest to oldest
            event_partition = get_partition(event)
            if event_partition not in latest_materialization_by_partition:
                latest_materialization_by_partition[event_partition] = event
            if len(latest_materialization_by_partition) == len(partitions):
                break

        # return materializations in the same order as the provided partitions, None if
        # materialization does not exist
        ordered_materializations = [
            latest_materialization_by_partition.get(partition) for partition in partitions
        ]

        return [
            GrapheneAssetMaterialization(event=event) if event else None
            for event in ordered_materializations
        ]


class GrapheneAssetNodeOrError(graphene.Union):
    class Meta:
        types = (GrapheneAssetNode, GrapheneAssetNotFoundError)
        name = "AssetNodeOrError"
